#define _GNU_SOURCE
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ptrace.h>
#include <sys/prctl.h>
#include <sys/syscall.h>
#include <sys/signal.h>
#include <linux/seccomp.h>
#include <linux/filter.h>

int main(void) {
  pid_t middle_child = fork();
  if (middle_child == -1)
    err(1, "fork");
  if (middle_child != 0) {
    sleep(3);
    exit(0);
  }

  pid_t child = fork();
  if (child == -1)
    err(1, "fork");
  if (child == 0) {
    // set some silly filter that blocks the chdir() syscall,
    // assuming that the syscall only makes x86-64 ABI syscalls.
    // not actually secure like this, just as an example...
    struct sock_filter filter[] = {
      BPF_STMT(BPF_LD | BPF_W | BPF_ABS, (offsetof(struct seccomp_data, nr))),
      BPF_JUMP(BPF_JMP | BPF_JEQ | BPF_K, __NR_chdir, 0, 1),
      BPF_STMT(BPF_RET | BPF_K, SECCOMP_RET_TRACE),
      BPF_STMT(BPF_RET | BPF_K, SECCOMP_RET_ALLOW),
    };
    struct sock_fprog prog = { .len = sizeof(filter)/sizeof(filter[0]), .filter = filter };
    if (prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0))
      err(1, "prctl");
    if (syscall(__NR_seccomp, SECCOMP_SET_MODE_FILTER, 0, &prog))
      err(1, "seccomp");
    sleep(1); // wait for tracer attachment

    errno = 0;
    printf("chdir('/') = %d (%m)\n", chdir("/"));
    errno = 0;
    printf("chdir('/') = %d (%m)\n", chdir("/"));
    exit(0);
  }

  if (ptrace(PTRACE_SEIZE, child, NULL, PTRACE_O_TRACESECCOMP))
    err(1, "PTRACE_SEIZE");

  sleep(2);
  exit(0);
}
